#include <iostream>
#include <fstream>
#include <sstream>
#include <memory>
#include <string>

#include <grpcpp/grpcpp.h>
#include <turbo-rpc/turbo.grpc.pb.h>

#include <pqxx/pqxx>



using namespace grpc;
using namespace Turbo;
using namespace std;

#define TURBODB "turbodb"
// Logic and data behind the server's behavior.
class TurboServiceImpl final : public TurboService::Service 
{
	public:

		TurboServiceImpl()
		{
			pqxx::connection conn{db_connect_options.c_str()};
			pqxx::work txn(conn);


			cout << "Creating turbod tables" << endl;
			txn.exec(turbod_schema.c_str());
			txn.commit();

		}

		Status Ping(ServerContext *context, const PingRequest_t *request, PingReply_t *reply) override 
		{
			return Status::OK;
		}

		Status BoostAdd(ServerContext *context, const BoostAddRequest_t *request, BoostAddReply_t *reply) override 
		{
			try
			{
				pqxx::connection conn{db_connect_options.c_str()};
				pqxx::work txn(conn);

				cout << "Requested creation of new boost named " << request->name() << endl;

				conn.prepare("boostcheck", "select boost_id from boosts where boost_name = $1");
				const auto check_res =  txn.prepared("boostcheck")(request->name()).exec();

				if(check_res.size() != 0)
				{
					return Status(ABORTED, "Boost already exists");
				}

				conn.prepare("boostadd", "insert into boosts (boost_name, create_time) values ($1, NOW()) returning boost_id ");
				const auto add_res =  txn.prepared("boostadd")(request->name()).exec();

				if(add_res.size() != 1)
				{
					return Status(ABORTED, "Unexpected sql reply.");
				}

				reply->set_boost_id(add_res[0]["boost_id"].as<int>());
				txn.commit();
				return Status::OK;
			}
			catch(const std::exception& e)
			{
				return Status(ABORTED, e.what());
			}

			
		}

		Status VersionAdd(ServerContext *context, const VersionAddRequest_t *request, VersionAddReply_t *reply) override 
		{
			try
			{
				pqxx::connection conn{db_connect_options.c_str()};
				pqxx::work txn(conn);

				cout << "Requested creation of new version for boost id " << request->boost_id() << endl;

				conn.prepare("boostcheck", "select boost_id,boost_name from boosts where boost_id = $1");
				const auto check_res =  txn.prepared("boostcheck")(request->boost_id()).exec();

				if(check_res.size() != 1)
				{
					return Status(ABORTED, "Boost id not found");
				}

				conn.prepare("versionadd", "insert into boost_versions ( boost_id, orig_binary, submit_time )"
						" values ($1, $2, NOW()) returning ver_id ");

				const auto boost_id    = check_res[0]["boost_id"].as<int>();
				const auto binary_data = pqxx::binarystring(request->version_contents().c_str(), request->version_contents().size());
				const auto add_res     = txn.prepared("versionadd")(boost_id)(binary_data).exec();

				if(add_res.size() != 1)
				{
					return Status(ABORTED, "Unexpected sql reply.");
				}

				reply->set_version_id(add_res[0]["ver_id"].as<int>());
				txn.commit();
				return Status::OK;
			}
			catch(const std::exception& e)
			{
				return Status(ABORTED, e.what());
			}

		}

	private:
		string db_connect_options = string("dbname=" TURBODB " host=localhost");
		const string turbod_schema = 
			R"""(
				-- avoid NOTICES if tables already exist.
				SET client_min_messages = error;

				-- avoid error if type already exists
				DO $$ BEGIN
					create type InputProvenance as ENUM ('User', 'AFL', 'QSym', 'Unknown');
				EXCEPTION
					when duplicate_object then null;
				END $$;

				create table if not exists boosts
				(
					boost_id            serial primary key,
					boost_name          bytea,
					create_time         timestamp
				);

				create table if not exists boost_versions
				(
					ver_id              serial primary key,
					boost_id            integer references boosts(boost_id),
					orig_binary         bytea,
					orig_bin_sha        bytea,
					zafl_binary         bytea,
					zafl_bin_sha        bytea,
					submit_time         timestamp,
					start_time          timestamp,
					end_time            timestamp,
					err                 bool
				);

				create table if not exists inputs
				(
					input_id            serial primary key,
					boost_id            integer references boosts(boost_id),
					create_time         timestamp,
					prov                InputProvenance,
					input_data          bytea,
					sha256              bytea
				);

				create table if not exists forensics
				(
					forensic_id         serial primary key,
					input_id            integer references inputs(input_id),
					ver_id              integer references boost_versions(ver_id),
					exit_code           integer,
					backtrace           bytea,
					prov                InputProvenance
				);

				create table if not exists seeds
				(
					seed_id             serial primary key,
					input_id            integer references inputs(input_id),
					prov                InputProvenance
				);


				create table if not exists zafl_logs
				(
					log_id              serial primary key,
					log                 bytea,
					ver_id              integer references boost_versions(ver_id)
				);

				create table if not exists fuzz_logs
				(
					log_id              serial primary key,
					log                 bytea,
					ver_id              integer references boost_versions(ver_id),
					prov                InputProvenance
				);
			)""";



};

void RunServer() 
{
	// listening port
	const auto server_address=string("0.0.0.0:55155");

	// create the service, build it, bind to listening port, and register.
	TurboServiceImpl service;
	ServerBuilder builder;
	builder.AddListeningPort(server_address, grpc::InsecureServerCredentials());
	builder.RegisterService(&service);

	// Finally start the server.
	auto server = builder.BuildAndStart();

	// log
	cout << "Server listening on " << server_address << endl;

	// Wait for the server to shutdown. Note that some other thread must be
	// responsible for shutting down the server for this call to ever return.
	server->Wait();
}


int main(int argc, char **argv) 
{
	const auto create_db_res = system("createdb " TURBODB);
	if(create_db_res == -1)
	{
		cout << "Could not create process for creating database" << endl;
		perror("main"); 
	}
	RunServer();
	return 0;
}
